from django.conf import settings
from django.db import models
from edc_crf.model_mixins import CrfModelMixin
from edc_model import models as edc_models

from ..model_mixins import SubjectVisitMissedModelMixin
from .subject_visit_missed_reasons import SubjectVisitMissedReasons


class SubjectVisitMissed(
    CrfModelMixin,
    SubjectVisitMissedModelMixin,
    edc_models.BaseUuidModel,
):

    subject_visit = models.OneToOneField(
        settings.SUBJECT_VISIT_MODEL,
        on_delete=models.PROTECT,
        related_name="edc_subject_visit",
    )

    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="missed_reasons"
    )

    class Meta(CrfModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Missed Visit Report"
        verbose_name_plural = "Missed Visit Report"



from django.db.models.signals import post_save
from django.dispatch import receiver


@receiver(post_save, weak=False, dispatch_uid="visit_tracking_check_in_progress_on_post_save")
def visit_tracking_check_in_progress_on_post_save(
    sender, instance, raw, created, using, update_fields, **kwargs  # noqa
):
    """Calls method on the visit tracking instance"""
    if not raw and not update_fields:
        try:
            instance.appointment
        except AttributeError:
            pass
        else:
            try:
                instance.check_appointment_in_progress()
            except AttributeError as e:
                if "check_appointment_in_progress" not in str(e):
                    raise



from django.contrib import admin

from ..admin_site import edc_visit_tracking_admin
from .models import SubjectVisit


@admin.register(SubjectVisit, site=edc_visit_tracking_admin)
class SubjectVisitAdmin(admin.ModelAdmin):
    pass



from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_constants.constants import (
    ALIVE,
    DEAD,
    HOSPITALIZED,
    NO,
    NOT_APPLICABLE,
    OTHER,
    YES,
)
from edc_facility.import_holidays import import_holidays
from edc_list_data import load_list_data
from edc_metadata.models import CrfMetadata
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule import Crf, FormsCollection, Schedule, Visit, VisitSchedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import MISSED_VISIT
from edc_visit_tracking.models import SubjectVisitMissedReasons

from ..forms import SubjectVisitMissedForm
from ..helper import Helper
from ..models import SubjectVisit, list_data


class TestVisit(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        load_list_data(
            list_data=list_data,
            model_name="edc_visit_tracking.subjectvisitmissedreasons",
        )
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        crfs = FormsCollection(
            Crf(show_order=1, model="edc_visit_tracking.crfone", required=True),
            Crf(show_order=2, model="edc_visit_tracking.crftwo", required=True),
            Crf(show_order=3, model="edc_visit_tracking.crfthree", required=True),
            Crf(show_order=4, model="edc_visit_tracking.crffour", required=True),
            Crf(show_order=5, model="edc_visit_tracking.crffive", required=True),
        )
        crfs_missed = FormsCollection(
            Crf(
                show_order=1,
                model="edc_visit_tracking.subjectvisitmissed",
                required=True,
            ),
        )

        visit_schedule1 = VisitSchedule(
            name="visit_schedule1",
            offstudy_model="edc_visit_tracking.subjectoffstudy",
            death_report_model="edc_visit_tracking.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule1 = Schedule(
            name="schedule1",
            onschedule_model="edc_visit_tracking.onscheduleone",
            offschedule_model="edc_visit_tracking.offscheduleone",
            consent_model="edc_visit_tracking.subjectconsent",
            appointment_model="edc_appointment.appointment",
        )
        visits = []
        for index in range(0, 4):
            visits.append(
                Visit(
                    code=f"{index + 1}000",
                    title=f"Day {index + 1}",
                    timepoint=index,
                    rbase=relativedelta(days=index),
                    rlower=relativedelta(days=0),
                    rupper=relativedelta(days=6),
                    requisitions=None,
                    crfs=crfs,
                    crfs_missed=crfs_missed,
                    allow_unscheduled=True,
                )
            )
        for visit in visits:
            schedule1.add_visit(visit)
        visit_schedule1.add_schedule(schedule1)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )

    @staticmethod
    def get_subject_visit():
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment, reason=MISSED_VISIT
        )
        return appointment, subject_visit

    @override_settings(
        SUBJECT_MISSED_VISIT_REASONS_MODEL="edc_visit_tracking.subjectvisitmissed"
    )
    def test_(self):
        self.helper.consent_and_put_on_schedule()
        appointment, subject_visit = self.get_subject_visit()
        opts = dict(
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            timepoint=appointment.timepoint,
        )
        self.assertGreater(CrfMetadata.objects.filter(**opts).count(), 0)
        subject_visit.reason = MISSED_VISIT
        subject_visit.save()
        self.assertEqual(1, CrfMetadata.objects.filter(**opts).count())

    def test_subject_visit_missed_form_survivial_and_ltfu(self):
        self.helper.consent_and_put_on_schedule()
        _, subject_visit = self.get_subject_visit()
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=DEAD,
            contact_attempted=YES,
            contact_attempts_count=4,
            contact_attempts_explained=None,
            contact_last_date=get_utcnow(),
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            contact_made=YES,
            ltfu=YES,
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("ltfu", form._errors)

        data.update(survival_status=DEAD, ltfu=NOT_APPLICABLE)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("ltfu", form._errors)

        data.update(survival_status=ALIVE, ltfu=NOT_APPLICABLE)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("ltfu", form._errors)

    def test_subject_visit_missed_form_missed_reasons(self):
        self.helper.consent_and_put_on_schedule()
        _, subject_visit = self.get_subject_visit()
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=ALIVE,
            contact_attempted=YES,
            contact_attempts_count=1,
            contact_last_date=get_utcnow(),
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            contact_made=YES,
            ltfu=NO,
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("missed_reasons_other", form._errors)
        data.update(missed_reasons=[SubjectVisitMissedReasons.objects.get(name=OTHER)])
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("missed_reasons_other", form._errors)

    def test_subject_visit_missed_form_attempts(self):
        self.helper.consent_and_put_on_schedule()
        _, subject_visit = self.get_subject_visit()
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=ALIVE,
            contact_attempted=NO,
            contact_made=NOT_APPLICABLE,
            contact_attempts_count=None,
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            ltfu=YES,
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("contact_attempts_count", form._errors)

        data.update(contact_attempted=YES, contact_made=NO, contact_attempts_count=None)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()

        self.assertIn("contact_attempts_count", form._errors)

        data.update(contact_attempted=YES, contact_made=NO, contact_attempts_count=2)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("contact_attempts_explained", form._errors)

        data.update(contact_attempted=YES, contact_made=NO, contact_attempts_count=3)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("contact_attempts_explained", form._errors)



from dateutil.relativedelta import relativedelta
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from edc_appointment.constants import INCOMPLETE_APPT
from edc_appointment.creators import UnscheduledAppointmentCreator
from edc_appointment.models import Appointment
from edc_facility.import_holidays import import_holidays
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED

from ..helper import Helper
from ..models import BadCrfOneInline, CrfOne, CrfOneInline, OtherModel, SubjectVisit
from ..visit_schedule import visit_schedule1, visit_schedule2


class TestVisit(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )

    def test_methods(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        instance = CrfOne(subject_visit=subject_visit)

        self.assertEqual(instance.subject_visit, subject_visit)
        self.assertEqual(instance.related_visit_model_attr(), "subject_visit")
        self.assertEqual(CrfOne.related_visit_model_attr(), "subject_visit")
        self.assertEqual(CrfOne.related_visit_model_cls(), SubjectVisit)

    def test_crf_related_visit_model_attrs(self):
        """Assert models using the CrfModelMixin can determine which
        attribute points to the visit model foreignkey.
        """
        self.assertEqual(CrfOne().related_visit_model_attr(), "subject_visit")
        self.assertEqual(CrfOne.objects.all().count(), 0)

    def test_crf_visit_model(self):
        """Assert models using the CrfModelMixin can determine which
        visit model is in use for the app_label.
        """
        self.assertEqual(CrfOne().related_visit_model_cls(), SubjectVisit)
        self.assertEqual(CrfOne.objects.all().count(), 0)

    def test_crf_inline_model_attrs(self):
        """Assert inline model can find visit instance from parent."""
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        other_model = OtherModel.objects.create()
        crf_one_inline = CrfOneInline.objects.create(crf_one=crf_one, other_model=other_model)
        self.assertEqual(crf_one_inline.related_visit.pk, subject_visit.pk)

    def test_crf_inline_model_parent_model(self):
        """Assert inline model cannot find parent, raises exception."""
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        other_model = OtherModel.objects.create()
        self.assertRaises(
            ImproperlyConfigured,
            BadCrfOneInline.objects.create,
            crf_one=crf_one,
            other_model=other_model,
        )

    def test_crf_inline_model_attrs2(self):
        """Assert inline model can find visit instance from parent."""
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        other_model = OtherModel.objects.create()
        crf_one_inline = CrfOneInline.objects.create(crf_one=crf_one, other_model=other_model)
        self.assertIsInstance(crf_one_inline.related_visit, SubjectVisit)

    def test_get_previous_model_instance(self):
        """Assert model can determine the previous."""
        self.helper.consent_and_put_on_schedule()
        for index, appointment in enumerate(
            Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        ):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED,
            )
        subject_visits = SubjectVisit.objects.all().order_by(
            "appointment__timepoint", "appointment__visit_code_sequence"
        )
        self.assertEqual(subject_visits.count(), 4)
        subject_visit = subject_visits[0]
        self.assertIsNone(subject_visit.previous_visit)
        subject_visit = subject_visits[1]
        self.assertEqual(subject_visit.previous_visit.pk, subject_visits[0].pk)
        subject_visit = subject_visits[2]
        self.assertEqual(subject_visit.previous_visit.pk, subject_visits[1].pk)
        subject_visit = subject_visits[3]
        self.assertEqual(subject_visit.previous_visit.pk, subject_visits[2].pk)

    def test_get_previous_model_instance2(self):
        """Assert model can determine the previous even when unscheduled
        appointment are inserted.
        """
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()

        appointment = appointments.last()
        for i in [1, 2]:
            appointment = UnscheduledAppointmentCreator(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=appointment.visit_schedule_name,
                schedule_name=appointment.schedule_name,
                visit_code=appointment.visit_code,
                visit_code_sequence=appointment.visit_code_sequence + 1,
                timepoint=appointment.timepoint,
                facility=appointment.facility,
            ).appointment
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=(appointment.appt_datetime + relativedelta(days=i)),
                reason=UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()

        subject_visits = SubjectVisit.objects.all().order_by(
            "appointment__timepoint", "appointment__visit_code_sequence"
        )
        self.assertEqual(subject_visits.count(), 6)

        subject_visit = subject_visits[0]
        self.assertIsNone(subject_visit.previous_visit)
        subject_visit = subject_visits[1]
        self.assertEqual(subject_visit.previous_visit, subject_visits[0])
        subject_visit = subject_visits[2]
        self.assertEqual(subject_visit.previous_visit, subject_visits[1])
        subject_visit = subject_visits[3]
        self.assertEqual(subject_visit.previous_visit, subject_visits[2])

    def test_missed_no_crfs(self):
        pass



from .crfs import VisitTrackingCrfModelMixin
from .requisitions import VisitTrackingRequisitionModelMixin
from .subject_visit_missed_model_mixin import SubjectVisitMissedModelMixin
from .utils import get_related_visit_model_attr
from .visit_model_mixin import (
    CaretakerFieldsMixin,
    PreviousVisitError,
    PreviousVisitModelMixin,
    VisitModelFieldsMixin,
    VisitModelMixin,
)



from datetime import datetime


class VisitTrackingModelFormMixin:
    report_datetime_field_attr = "report_datetime"

    @property
    def subject_identifier(self) -> str:
        return self.cleaned_data.get("subject_identifier") or self.instance.subject_identifier

    @property
    def report_datetime(self) -> datetime:
        return self.cleaned_data.get(self.report_datetime_field_attr) or getattr(
            self.instance, self.report_datetime_field_attr
        )



from django.apps import apps as django_apps
from edc_appointment.managers import AppointmentManager
from edc_appointment.model_mixins import AppointmentModelMixin
from edc_list_data.model_mixins import ListModelManager, ListModelMixin
from edc_model.models import HistoricalRecords

from ..managers import CrfModelManager, VisitModelManager
from ..model_mixins import VisitModelMixin, VisitTrackingCrfModelMixin


class ManagerTestCase:
    app_label = None

    appointment_manager_cls = AppointmentManager
    appointment_model_cls = AppointmentModelMixin
    crf_manager_cls = CrfModelManager
    crf_model_cls = VisitTrackingCrfModelMixin
    history_manager_cls = HistoricalRecords
    list_manager_cls = ListModelManager
    list_model_cls = ListModelMixin
    visit_manager_cls = VisitModelManager
    visit_model_cls = VisitModelMixin
    requisition_manager_cls = VisitModelManager
    requisition_model_cls = VisitModelMixin

    @property
    def appointment_model(self):
        """Returns the appointment model name."""
        appointment_model = None
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if issubclass(model, (self.appointment_model_cls,)):
                appointment_model = model._meta.label_lower
                break
        return appointment_model

    @property
    def visit_models(self):
        """Returns a list of "Visit" model names."""
        visit_models = []
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if issubclass(model, (self.visit_model_cls,)):
                visit_models.append(model._meta.label_lower)
        return visit_models

    @property
    def requisition_models(self):
        """Returns a list of "Requisition" model names."""
        requisition_models = []
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if issubclass(model, (self.requisition_model_cls,)):
                requisition_models.append(model._meta.label_lower)
        return requisition_models

    @property
    def list_models(self):
        """Returns a list of "List" model names."""
        list_models = []
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if issubclass(model, (self.list_model_cls,)):
                list_models.append(model._meta.label_lower)
        return list_models

    @property
    def non_crf_models(self):
        """Returns a list of non-CRF model names.

        Note: A non-crf model is does not include appointment,
        visit, or list models.
        """
        non_crf_models = []
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            classes = (
                self.crf_model_cls,
                self.list_model_cls,
                self.appointment_model_cls,
                self.visit_manager_cls,
            )
            if not issubclass(model, classes):
                non_crf_models.append(model._meta.label_lower)
        return non_crf_models

    @property
    def crf_models(self):
        """Returns a list of CRF model names."""
        crf_models = []
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if issubclass(model, (self.crf_model_cls,)) and not issubclass(
                model, (self.requisition_model_cls,)
            ):
                crf_models.append(model._meta.label_lower)
        return crf_models

    def have_manager_or_raise(self, models=None, manager_cls=None):
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if model._meta.label_lower in models:
                if (
                    model._default_manager.__class__ != manager_cls
                    and "historical" not in model._meta.label_lower
                ):
                    self.assertTrue(
                        issubclass(model._default_manager.__class__, manager_cls),
                        msg=(
                            f"model {model._meta.label_lower}. "
                            f"Got manager {model._default_manager.__class__} "
                            f"is not a subclass of {manager_cls}."
                        ),
                    )

    def have_history_or_raise(self, models=None):
        app_config = django_apps.get_app_config(self.app_label)
        for model in app_config.get_models():
            if model._meta.label_lower in models:
                if "historical" not in model._meta.label_lower:
                    self.assertTrue(
                        issubclass(model.history.__class__, self.history_manager_cls),
                        msg=(
                            f"model {model._meta.label_lower}. "
                            f"Got history manager {model.history.__class__} "
                            f"is not a subclass of {self.history_manager_cls}."
                        ),
                    )

    def test_crf_default_manager_subclass(self):
        self.have_manager_or_raise(models=self.crf_models, manager_cls=self.crf_manager_cls)

    def test_crf_history_manager_subclass(self):
        self.have_history_or_raise(models=self.crf_models)

    def test_requisition_default_manager_subclass(self):
        self.have_manager_or_raise(
            models=self.requisition_models, manager_cls=self.requisition_manager_cls
        )

    def test_requisition_history_manager_subclass(self):
        self.have_history_or_raise(models=self.requisition_models)



from edc_list_data.model_mixins import ListModelMixin


class SubjectVisitMissedReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Subject Missed Visit Reasons"
        verbose_name_plural = "Subject Missed Visit Reasons"



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from edc_appointment.choices import APPT_TIMING
from edc_appointment.constants import MISSED_APPT, SCHEDULED_APPT, UNSCHEDULED_APPT
from edc_constants.constants import INCOMPLETE
from edc_metadata.metadata_helper import MetadataHelperMixin
from edc_visit_schedule.utils import is_baseline

from .constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from .utils import get_subject_visit_missed_model_cls

if TYPE_CHECKING:
    from edc_appointment.models import Appointment


class SubjectVisitReasonUpdaterError(Exception):
    pass


class SubjectVisitReasonUpdaterBaselineError(Exception):
    pass


class SubjectVisitReasonUpdaterCrfsExistsError(Exception):
    pass


class SubjectVisitReasonUpdaterRequisitionsExistsError(Exception):
    pass


class SubjectVisitReasonUpdaterMissedVisitNotAllowed(Exception):
    pass


class SubjectVisitReasonUpdater(MetadataHelperMixin):
    """A class to try to update `reason` field based on the
    response from appointment timing and appointment reason.
    """

    metadata_helper_instance_attr: str = "appointment"

    def __init__(
        self,
        appointment: Appointment = None,
        appt_timing: str = None,
        appt_reason: str = None,
        commit: bool | None = None,
    ):
        self.appointment = appointment
        if not getattr(self.appointment, "id", None):
            raise SubjectVisitReasonUpdaterError(
                "Appointment instance must exist. Got `id` is None"
            )
        self.commit = commit
        self.appt_timing = appt_timing or self.appointment.appt_timing
        if self.appt_timing not in [a for a, b in APPT_TIMING]:
            raise SubjectVisitReasonUpdaterError(
                f"Invalid value for appt_timing. "
                f"Expected on of {[a for a, b in APPT_TIMING]}. Got {self.appt_timing}"
            )
        self.appt_reason = appt_reason or self.appointment.appt_reason
        try:
            self.related_visit = getattr(
                self.appointment, self.appointment.related_visit_model_attr()
            )
        except ObjectDoesNotExist:
            self.related_visit = None
        except AttributeError as e:
            if "related_visit_model_attr" not in str(e):
                raise
            self.related_visit = None

    def update_or_raise(self) -> None:
        if self.appt_timing == MISSED_APPT and is_baseline(instance=self.appointment):
            raise SubjectVisitReasonUpdaterBaselineError(
                "Invalid. A baseline appointment may not be reported `missed`"
            )
        elif self.appt_timing == MISSED_APPT and self.appt_reason in [
            SCHEDULED_APPT,
            UNSCHEDULED_APPT,
        ]:
            self._update_visit_to_missed_or_raise()
        elif self.appt_timing != MISSED_APPT and self.appt_reason in [
            SCHEDULED_APPT,
            UNSCHEDULED_APPT,
        ]:
            self._update_visit_to_not_missed_or_raise()
        else:
            raise SubjectVisitReasonUpdaterError(
                f"Condition not handled. Got appt_reason={self.appt_reason}, "
                f"appt_timing={self.appt_timing}"
            )

    def _update_visit_to_missed_or_raise(self) -> None:
        self.missed_visit_allowed_or_raise()
        if self.related_visit:
            self.related_visit.reason = MISSED_VISIT
            self.related_visit.document_status = INCOMPLETE
            if self.commit:
                self.related_visit.save_base(update_fields=["reason", "document_status"])
                self.related_visit.refresh_from_db()

    def _update_visit_to_not_missed_or_raise(self) -> None:
        """Updates the subject visit instance from MISSED_VISIT
        to SCHEDULED or UNSCHEDULED.
        """
        if self.related_visit:
            reason = self.get_reason_from_appt_reason(self.appt_reason)
            self.delete_subject_visit_missed_if_exists()
            self.related_visit.reason = reason
            self.related_visit.document_status = INCOMPLETE
            if self.related_visit.comments:
                self.related_visit.comments = self.related_visit.comments.replace(
                    "[auto-created]", ""
                )
            if self.commit:
                self.related_visit.save_base(
                    update_fields=["reason", "document_status", "comments"]
                )
                self.related_visit.refresh_from_db()

    def missed_visit_allowed_or_raise(self) -> None:
        """Raises an exception if not allowed.

        May be not allowed by settings attr or
        if CRF/Requisition metadata already exists.

        See also: EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED
        """
        if (
            self.appointment.visit_code_sequence > 0
            and not self.allow_missed_unscheduled_appts
        ):
            raise SubjectVisitReasonUpdaterMissedVisitNotAllowed(
                "Invalid. An unscheduled appointment may not be reported `missed`."
                "Try to cancel the appointment instead. "
            )
        else:
            self.raise_if_keyed_data_exists()

    def raise_if_keyed_data_exists(self) -> None:
        """Raises an exception if CRF or Requisition metadata exists"""
        if self.crf_metadata_keyed_exists:
            raise SubjectVisitReasonUpdaterCrfsExistsError(
                "Invalid. CRFs have already been entered for this timepoint."
            )
        elif self.requisition_metadata_keyed_exists:
            raise SubjectVisitReasonUpdaterRequisitionsExistsError(
                "Invalid. Requisitions have already been entered " "for this timepoint."
            )

    def delete_subject_visit_missed_if_exists(self) -> None:
        """Deletes the subject visit missed report if it exists"""
        get_subject_visit_missed_model_cls().objects.filter(
            subject_visit__appointment=self.appointment
        ).delete()

    def get_reason_from_appt_reason(self, appt_reason: str) -> str:
        """Returns a subject visit reason given the appt reason"""
        if appt_reason == SCHEDULED_APPT:
            visit_reason = SCHEDULED
        elif appt_reason == UNSCHEDULED_APPT and self.appointment.visit_code_sequence > 0:
            visit_reason = UNSCHEDULED
        else:
            raise SubjectVisitReasonUpdaterError(
                "Update failed. This is not an unscheduled appointment. "
                f"Got visit_code_sequence > 0 for appt_reason={appt_reason}"
            )
        return visit_reason

    @property
    def allow_missed_unscheduled_appts(self) -> bool:
        """Returns value of settings attr or False"""
        return getattr(settings, "EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED", False)



from edc_list_data.model_mixins import ListModelMixin


class VisitReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Visit Reasons"
        verbose_name_plural = "Visit Reasons"



from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from typing import TYPE_CHECKING, Any, List, Optional
from zoneinfo import ZoneInfo

from django import forms
from django.conf import settings
from edc_appointment.constants import MISSED_APPT
from edc_appointment.form_validators import (
    WindowPeriodFormValidatorMixin,
    validate_appt_datetime_unique,
)
from edc_constants.constants import OTHER
from edc_form_validators import INVALID_ERROR, REQUIRED_ERROR, FormValidator
from edc_metadata.constants import KEYED
from edc_metadata.utils import (
    get_crf_metadata_model_cls,
    get_requisition_metadata_model_cls,
)
from edc_utils import formatted_datetime
from edc_visit_schedule.utils import is_baseline

from ..constants import MISSED_VISIT, UNSCHEDULED
from ..utils import get_subject_visit_missed_model_cls
from ..visit_sequence import VisitSequence, VisitSequenceError

if TYPE_CHECKING:
    from edc_appointment.models import Appointment

EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED = getattr(
    settings, "EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED", False
)


class VisitFormValidator(WindowPeriodFormValidatorMixin, FormValidator):

    """Form validator for visit models (e.g. subject_visit).

    See also `report_datetime` checks in the
    `VisitTrackingCrfModelFormMixin`.
    """

    visit_sequence_cls = VisitSequence
    validate_missed_visit_reason = True
    validate_unscheduled_visit_reason = True
    report_datetime_field_attr = "report_datetime"

    def _clean(self) -> None:
        super()._clean()
        if not self.appointment:
            raise forms.ValidationError(
                {"appointment": "This field is required"}, code=REQUIRED_ERROR
            )

        validate_appt_datetime_unique(
            form_validator=self,
            appointment=self.appointment,
            appt_datetime=self.appointment.appt_datetime,
            form_field="appointment",
        )

        self.validate_visit_datetime_unique()

        self.validate_visit_datetime_not_before_appointment()

        self.validate_visit_datetime_matches_appt_datetime_at_baseline()

        self.validate_visit_datetime_in_window_period()

        self.validate_visits_completed_in_order()

        self.validate_visit_code_sequence_and_reason()

        self.validate_visit_reason()

        self.required_if(OTHER, field="info_source", field_required="info_source_other")

    @property
    def subject_identifier(self) -> str:
        return self.appointment.subject_identifier

    @property
    def appointment(self) -> Appointment:
        appointment = None
        if "appointment" in self.cleaned_data:
            appointment = self.cleaned_data.get("appointment")
        elif self.instance:
            appointment = self.instance.appointment
        if not appointment:
            self.raise_validation_error("Appointment is required.")
        return appointment

    @property
    def report_datetime(self) -> datetime | None:
        """Returns report datetime in local timezone (from form
        cleaned_data).
        """
        report_datetime = None
        if "report_datetime" in self.cleaned_data:
            report_datetime = self.cleaned_data.get("report_datetime")
        elif self.instance:
            report_datetime = self.instance.report_datetime
        return report_datetime

    @property
    def report_datetime_utc(self) -> datetime | None:
        """Returns report datetime in UTC timezone"""
        if self.report_datetime:
            return self.report_datetime.astimezone(ZoneInfo("UTC"))
        return None

    @property
    def appt_datetime_local(self) -> datetime:
        """Returns appt datetime in local timezone"""
        return self.appointment.appt_datetime.astimezone(ZoneInfo(settings.TIME_ZONE))

    def validate_visit_datetime_in_window_period(self, *args) -> None:
        """Asserts the report_datetime is within the visits lower and
        upper boundaries of the visit_schedule.schdule.visit.

        See also `edc_visit_schedule`.
        """
        if self.report_datetime:
            args = [self.appointment, self.report_datetime, self.report_datetime_field_attr]
            self.datetime_in_window_or_raise(*args)

    def validate_visit_datetime_unique(self: Any) -> None:
        """Assert one visit report per day"""
        if self.report_datetime:
            qs = self.instance.__class__.objects.filter(
                subject_identifier=self.subject_identifier,
                report_datetime__date=self.report_datetime_utc.date(),
                visit_schedule_name=self.instance.visit_schedule_name,
                schedule_name=self.instance.schedule_name,
            )
            if getattr(self.instance, "id"):
                qs = qs.exclude(id=self.instance.id)
            if qs.count() > 1:
                raise self.raise_validation_error(
                    {"report_datetime": "Visit report already exist for this date (M)"},
                    INVALID_ERROR,
                )
            elif qs.count() == 1:
                raise self.raise_validation_error(
                    {
                        "report_datetime": "A visit report already exists for this date. "
                        f"See {qs[0].visit_code}.{qs[0].visit_code_sequence}"
                    },
                    INVALID_ERROR,
                )

    def validate_visit_datetime_not_before_appointment(
        self,
    ) -> None:
        """Asserts the report_datetime is not before the
        appt_datetime.
        """
        if report_datetime_local := self.report_datetime:
            if report_datetime_local.date() < self.appt_datetime_local.date():
                appt_datetime_str = formatted_datetime(
                    self.appt_datetime_local, format_as_date=True
                )
                self.raise_validation_error(
                    {
                        "report_datetime": (
                            "Invalid. Cannot be before appointment date. "
                            f"Got appointment date {appt_datetime_str}"
                        )
                    },
                    INVALID_ERROR,
                )

    def validate_visit_datetime_matches_appt_datetime_at_baseline(self) -> None:
        """Asserts the report_datetime matches the appt_datetime
        as baseline.
        """
        if is_baseline(instance=self.appointment):
            if report_datetime_local := self.report_datetime:
                if report_datetime_local.date() != self.appt_datetime_local.date():
                    appt_datetime_str = formatted_datetime(
                        self.appt_datetime_local, format_as_date=True
                    )
                    self.raise_validation_error(
                        {
                            "report_datetime": (
                                "Invalid. Must match appointment date at baseline. "
                                "If necessary, change the appointment date and try again. "
                                f"Got appointment date {appt_datetime_str}"
                            )
                        },
                        INVALID_ERROR,
                    )

    def validate_visits_completed_in_order(self) -> None:
        """Asserts visits are completed in order."""
        visit_sequence = self.visit_sequence_cls(appointment=self.appointment)
        try:
            visit_sequence.enforce_sequence()
        except VisitSequenceError as e:
            raise forms.ValidationError(e, code=INVALID_ERROR)

    def validate_visit_code_sequence_and_reason(self) -> None:
        """Asserts the `reason` makes sense relative to the
        visit_code_sequence coming from the appointment.
        """
        appointment = self.appointment
        reason = self.cleaned_data.get("reason")
        if appointment:
            if not appointment.visit_code_sequence and reason == UNSCHEDULED:
                raise forms.ValidationError(
                    {"reason": "Invalid. This is not an unscheduled visit. See appointment."},
                    code=INVALID_ERROR,
                )
            if (
                appointment.visit_code_sequence
                and reason != UNSCHEDULED
                and EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED is False
            ):
                raise forms.ValidationError(
                    {"reason": "Invalid. This is an unscheduled visit. See appointment."},
                    code=INVALID_ERROR,
                )
            # raise if CRF metadata exist
            if reason == MISSED_VISIT and self.metadata_exists_for(
                entry_status=KEYED,
                exclude_models=[get_subject_visit_missed_model_cls()._meta.label_lower],
            ):
                raise forms.ValidationError(
                    {"reason": "Invalid. Some CRF data has already been submitted."},
                    code=INVALID_ERROR,
                )

    def validate_visit_reason(self) -> None:
        """Asserts that reason=missed if appointment is missed"""
        if (
            self.appointment.appt_timing == MISSED_APPT
            and self.cleaned_data.get("reason") != MISSED_VISIT
        ):
            self.raise_validation_error(
                {"reason": "Invalid. This appointment was reported as missed"}, INVALID_ERROR
            )

        if self.validate_missed_visit_reason:
            self.required_if(MISSED_VISIT, field="reason", field_required="reason_missed")

            self.required_if(
                OTHER, field="reason_missed", field_required="reason_missed_other"
            )

        if self.validate_unscheduled_visit_reason:
            if "reason_unscheduled" in self.cleaned_data:
                self.applicable_if(
                    UNSCHEDULED, field="reason", field_applicable="reason_unscheduled"
                )

                self.required_if(
                    OTHER,
                    field="reason_unscheduled",
                    field_required="reason_unscheduled_other",
                )

    def metadata_exists_for(
        self,
        entry_status: str = None,
        filter_models: Optional[List[str]] = None,
        exclude_models: Optional[List[str]] = None,
    ) -> int:
        """Returns True if metadata exists for this visit for
        the given entry_status.
        """
        exclude_opts: dict = {}
        filter_opts = deepcopy(self.crf_filter_options)
        filter_opts.update(entry_status=entry_status or KEYED)
        if filter_models:
            filter_opts.update(model__in=filter_models)
        if exclude_models:
            exclude_opts.update(model__in=exclude_models)
        return (
            get_crf_metadata_model_cls()
            .objects.filter(**filter_opts)
            .exclude(**exclude_opts)
            .count()
            + get_requisition_metadata_model_cls()
            .objects.filter(**filter_opts)
            .exclude(**exclude_opts)
            .count()
        )

    @property
    def crf_filter_options(self) -> dict:
        """Returns a dictionary of `filter` options when querying
        models CrfMetadata / RequisitionMetadata.
        """
        return dict(
            subject_identifier=self.subject_identifier,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            entry_status=KEYED,
        )
